# -*- coding: utf-8 -*-
#
# WebFilterAdmin a web configurator for dansguardian & squid.
# Copyright (C) 2005 Marco Andreini
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Dansguardian config file tool-classes

"""

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
#from BTrees.OIBTree import OIBTree
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree
from ZODB.POSException import ConflictError
from OFS.Folder import Folder
#from urllib import quote
from urlparse import urlparse
from utils import *

import re
import os.path
import fcntl


rInclude = re.compile(r".Include<([^>]+)>")


class BaseConfigFile(SimpleItem):
    """config file"""

    meta_type = "Config File"
    _v_mtime = 0
    _v_data = None
    security = ClassSecurityInfo()    

    def __init__(self, id, title, filepath):
        self.id = id
        self.title = title
        
        self.filepath = filepath
        self.getData()

    security.declarePrivate("real_read")
    def real_read(self, stream):
        raise NotImplementedError

    security.declarePrivate("real_write")
    def real_write(self, stream, data):
        raise NotImplementedError
    
    security.declarePrivate("getData")
    def getData(self):
        """read file (if need)"""
        
        # WARNING: pre-lock
        mtime = os.path.getmtime(self.filepath)
        
        if mtime > self._v_mtime or self._v_data is None:
            self._v_mtime = mtime
            
            datafile = open(self.filepath, "r")
            try:
                fcntl.lockf(datafile, fcntl.LOCK_SH | fcntl.LOCK_NB)
            except IOError:
                # FIXME: temporary unavailable, raise other exception.
                raise ConflictError, "lockf %s" % self.filepath
            try:
                self._v_data = self.real_read(datafile)
            finally:
                fcntl.lockf(datafile, fcntl.LOCK_UN)
                datafile.close()
        
        return self._v_data

    security.declarePrivate("writeData")
    def writeData(self):
        """write data"""
        
        # discard old data
        self._v_mtime = 0
        
        datafile = open(self.filepath, "w")
        try:
            fcntl.lockf(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            # FIXME: temporary unavailable
            raise ConflictError, "lockf %s" % self.filepath
        try:
            self.real_write(datafile, self._v_data)
        finally:
            fcntl.lockf(datafile, fcntl.LOCK_UN)
            datafile.close()

        # WARNING: post-lock
        self._v_mtime = os.path.getmtime(self.filepath)

        # FIXME: use the real path, don't use the acquisition.
        self.dansguardianReload()


class MessagesFile(BaseConfigFile):
    """messages file (dansguardian)"""
    
    meta_type = "Messages File (dansguardian)"
    icon = "wfa_messages.png"
    security = ClassSecurityInfo()

    def __init__(self, id, title, language, path):
        BaseConfigFile.__init__(self, id, title,
                                os.path.join(path, language, "messages"))
    
    def real_read(self, stream):
        """parsing messages file"""
        
        data = IOBTree()
        
        for line in stream.readlines():
            if line.startswith("#"):
                continue
            
            line = line.strip()
            if not line:
                continue
            
            num, msg = eval(line, {})
            num = int(num)
            if msg.strip().endswith(":"):
                msg += "..."
            data[num] = msg
        return data

    security.declareProtected(ManageWebFilter, "getMessage")
    def getMessage(self, reason):
        """get message from (int) reason"""
        
        data = self.getData()
        try:
            return data[reason]
        except KeyError:
            return "-"


InitializeClass(MessagesFile)


class PhraseListFile(BaseConfigFile, Folder):
    """phrase list file"""
    
    meta_type = "Phrase List File"
    icon = "wfa_phraselist.png"
    
    def real_read(self, stream):
        """read/parsing file-content"""
        
        pass
    
    def real_write(self, stream):
        
        pass


InitializeClass(PhraseListFile)


class WeightedPhraseListFile(BaseConfigFile, Folder):
    """weighted phrase list file"""
    
    meta_type = "Weighted Phrase List File"
    icon = "wfa_phraselist.png"

    def real_read(self, stream):
        """read/parsing file-content"""
        
        data = OOBTree()
        for line in stream.xreadlines():
            line = line.strip()
            if not line:
                continue
            
            if line.startswith("#"):
                continue
            
            match = rInclude.match(line)
            if match:
                filepath = match.group(1)
                id = filepath.replace("/", "_")
                if id.startswith("_"):
                    id = id[1:]
                if id in self.objectIds():
                    # remove & re-create?
                    continue
                self._setObject(id, self.__class__(id, filepath, filepath))
                continue
            items = line.split("#", 1)
            if len(items) > 1:
                data[items[0].strip()] = items[1].strip()
            else:
                data[items[0]] = None
        return data


InitializeClass(WeightedPhraseListFile)


class BaseListFile(BaseConfigFile, Folder):
    """list file"""
    
    meta_type = "List File"""
    icon = "wfa_baselist.png"

    security = ClassSecurityInfo()

    def __init__(self, id, title, path, parms=None):
        BaseConfigFile.__init__(self, id, title, path)
        try:
            self.__rmatch = parms["rmatch"]
            self.__format = parms["format"]
            self.__address = parms["address"]
        except KeyError:
            # FIXME: set __ attributes...
            pass
        except TypeError:
            pass

    def real_read(self, stream):
        """read/parsing file-content"""
        
        data = OOBTree()
        # or xreadlines?
        for line in stream.readlines():
            line = line.strip()
            if not line:
                continue
            
            if line.startswith("#"):
                continue
            
            match = rInclude.match(line)
            if match:
                filepath = match.group(1)
                id = filepath.replace("/", "_")
                if id.startswith("_"):
                    id = id[1:]
                if id in self.objectIds():
                    # remove & re-create?
                    continue
                self._setObject(id, self.__class__(id, filepath, filepath))
                continue
            items = line.split("#", 1)
            if len(items) > 1:
                data[items[0].strip()] = items[1].strip()
            else:
                data[items[0]] = None
        return data

    def real_write(self, stream, data):
        """real write file"""

        output = []
        for value in self.objectValues():
            output.append(".Include<%s>" % (value.filepath, ))
        for key, value in data.items():
            if value:
                output.append("%s # %s" % (key, value))
            else:
                output.append(key)
        output.append("")
        stream.write("\n".join(output))
    
    security.declareProtected(ManageWebFilter, "index_html")
    def index_html(self, REQUEST=None):
        """index_html"""
        
        return self.updateList_form(REQUEST)
    
    # specific to list-file

    security.declarePublic("isAddress")
    def isAddress(self):
        """is address"""

        return self.__address
    
    security.declareProtected(ManageWebFilter, "getList")
    def getList(self):
        """get list file"""
        
        return self.getData().items()

    security.declareProtected(ManageWebFilter, "has")
    def hasItem(self, item):
        """has this item"""
        
        return self.getData().has_key(item)

    security.declareProtected(ManageWebFilter, "remove")
    def remove(self, items):
        """remove datas"""
        
        data = self.getData()
        
        # WARNING: out of locking
        if not isinstance(items, (tuple, list)):
            items = (items, )
        
        for item in items:
            item = item.strip()
            if data.has_key(item):
                del data[item]
        
        self.writeData()
    
    security.declareProtected(ManageWebFilter, "add")
    def addItem(self, items):
        """add datas"""
        
        data = self.getData()
        
        # WARNING: out of locking
        if not isinstance(items, (tuple, list)):
            items = (items, )
        
        for item in items:
            item = item.strip()
            if item:
                data[item] = None
        
        self.writeData()
    
    # checks:
    
    security.declareProtected(ManageWebFilter, "alreadypresentList")
    def alreadypresentList(self, items):
        """check if url is defined elsewere"""

        # check all parent children (sons) with the same type.
        for oList in self.aq_inner.aq_parent.objectValues(self.meta_type):
            for item in items:
                if oList.hasItem(item):
                    return "\"%s\" è già presente nella lista \"%s.\"" % \
                           (item, oList.title)
        return False

    security.declareProtected(ManageWebFilter, "validateList")
    def validateList(self, items):
        """input string validation"""
        
        for item in items:
            item = item.strip()
            
            # for empty submissions
            if not item:
                continue

            if self.isAddress() and item.startswith("http://"):
                item = item[len("http://"):]
            if not self.__rmatch is None:
                if not self.__rmatch.match(item):
                    return "Errore nel formato di \"%s\". " \
                           "Deve essere nella forma \"%s\"" % \
                           (item, self.__format)
        return self.alreadypresentList(items)

    security.declareProtected(ManageWebFilter, "addList")
    def addList(self, addUrls=(), REQUEST=None):
        """add to list"""

        if not addUrls and not REQUEST is None:
            return self.updateList_form()
        
        err = self.validateList(addUrls)
        if err:
            if not REQUEST is None:
                REQUEST.set("errors", err)
                REQUEST.set("urls", "\n".join(addUrls))
            else:
                return err
        else:
            self.addItem(addUrls)
            if not REQUEST is None:
                REQUEST.set("message", "Gli elementi sono stati aggiunti.")
            else:
                return False
        return self.updateList_form()

    security.declareProtected(ManageWebFilter, "removeList")
    def removeList(self, removeUrls=(), REQUEST=None):
        """remove from list"""

        if not removeUrls and not REQUEST is None:
            return self.updateList_form(message="Selezionare almeno un elemento.")

        self.remove(removeUrls)
        return self.updateList_form(message="Gli elementi sono stati rimossi.")
    
    security.declarePrivate("addNotifiedList")
    def addNotifiedList(self, ids):
        """Add selected notified urls to site list"""
        
        requestIds = ", ".join(ids)
        
        sites = []
        for url in self.requestedUrls(requestid=requestIds):
            site = urlparse(url[0])[1]
            if not site:
                continue
            if not site in sites:
                sites.append(site)
        return self.addList(addUrls=sites)


InitializeClass(BaseListFile)    


item_hostname = dict(
    address=True,
    format="host.name.it",
    rmatch=re.compile(r"[a-z0-9][-.a-z0-9]*$")) # match lower case only

item_url = dict(
    address=True,
    format="host.name.it/urlpath.jpg",
    rmatch=re.compile(r"[a-z0-9][-.a-z0-9]*/[-/.,:;?%&$=+_a-zA-Z0-9]*$"))

item_ip = dict(
    address=True,
    format="10.0.1.2",
    rmatch=re.compile(r"[0-9]{1,3}(\.[0-9]{1,3}){3}$"))

item_mimetype = dict(
    address=False,
    format="text/plain",
    rmatch=re.compile(r"(?i)[-.a-z0-9]+/[-.a-z0-9]+$")) # ignore case

item_extension = dict(
    address=False,
    format=".exe",
    rmatch=re.compile(r"\.[a-z0-9]+$"))


class MainConfigFile(Folder):
    """main config file (dansguardian)"""

    meta_type = "Main Config File (dansguardian)"
    security = ClassSecurityInfo()
    _parameters = {
        # Content filtering files location
        # dansguardian.conf
        #bannedphraselist: 
        #weightedphraselist: 
        #exceptionphraselist: 
        "bannedsitelist": (BaseListFile, "Siti bloccati", item_hostname),
        # "greysitelist":
        "exceptionsitelist": (BaseListFile, "Siti permessi", item_hostname),
        "bannedurllist": (BaseListFile, "URL bloccati", item_url),
        #"greyurllist": (BaseListFile, "URL grigi?", item_url),
        "exceptionurllist": (BaseListFile, "URL permessi", item_url),
        "bannedregexpurllist": (BaseListFile, "Regexp URL bloccati"),
        "bannedextensionlist": (BaseListFile, "Estensioni bloccate",
                                item_extension),
        "bannedmimetypelist": (BaseListFile, "MIME Type bloccati",
                               item_mimetype),
        #picsfile: "/etc/dansguardian/pics"
        #contentregexplist: "/etc/dansguardian/contentregexplist"
        # dansguardianf1.conf
        "bannediplist": (BaseListFile, "IP bloccati", item_ip),
        "exceptioniplist": (BaseListFile, "IP permessi", item_ip),
        #"banneduserlist": BaseListFile,
        #"exceptionuserlist": BaseListFile,
        # language to use from languagedir
        "language": (MessagesFile, "Messaggi", "languagedir"),
        }
    
    def __init__(self, id, title, filepath1, filepath2):
        Folder.__init__(self, id)
        self._conf = filepath1
        self._f1conf = filepath2
        self.readConfigFiles()

    security.declareProtected(ManageWebFilter, "readConfigFiles")
    def readConfigFiles(self, REQUEST=None):
        """parsing main config file"""

        configuration = {}
        for fpath in (self._conf, self._f1conf):
            fdata = open(fpath, "r")
            for line in fdata.readlines():
                if line.startswith("#"):
                    continue
                line = line.strip()
                if not line:
                    continue
                try:
                    key, value = line.split("=", 1)
                except ValueError:
                    # ignore syntax error
                    continue

                key = key.strip()
                value = value.strip()
                if value.startswith("'") and value.endswith("'"):
                    value = value[1:-1]
                configuration[key] = value
            fdata.close()
        
        for key, values in self._parameters.items():
            klass, title = values[:2]
            if not configuration.has_key(key):
                # log ?
                raise Exception, "error on configuration %r" % (key, )
                # continue ?
            args = values[2:]
            values = []
            for arg in args:
                if isinstance(arg, basestring) and configuration.has_key(arg):
                    values.append(configuration[arg])
                else:
                    values.append(arg)
            try:
                self._delObject(key)
            except AttributeError:
                pass
            self._setObject(key, klass(key, title, configuration[key],
                                       *values))
            
        if not REQUEST is None:
            return "Instance configuration files successfully recreated."
            
    
InitializeClass(MainConfigFile)
