# -*- coding: utf-8 -*-
#
# WebFilterAdmin a web configurator for dansguardian & squid.
# Copyright (C) 2005 Marco Andreini
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
ProxyAdmin - WebFilterAdmin

"""

from Globals import InitializeClass, package_home
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from zExceptions import BadRequest
from DateTime import DateTime

from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.DirectoryView import addDirectoryViews
from Products.CMFCore.utils import getToolByName
from Products.CMFDefault.Portal import CMFSite, PortalGenerator

from ConfigFile import MainConfigFile
from Services import Service, systemService
from utils import *

from socket import getfqdn

import os
import os.path


class MenuItem:
    """menu item"""
    
    subMenu = False
    
    def __init__(self, title, action):
        self.title = title
        self._action = action
    
    def url(self, base):
        """base url"""
        
        return "%s/%s" % (base, self._action)


class SubMenuItem(MenuItem):
    """sub menu item"""
    
    subMenu = True


class Menu:
    """menu"""
    
    def __init__(self, *args):
        items = []
        
        for arg in args:
            title, action = arg[:2]
            items.append(MenuItem(title, action))
            for item in arg[2:]:
                items.append(SubMenuItem(*item))
        self._items = items

    def __getitem__(self, i):
        """getitem"""
        
        return self._items[i]

    def __len__(self):
        """len"""

        return len(self._items)

    def getItems(self, base):
        """all items"""

        return [{"title": item.title,
                 "url": item.url(base),
                 "subMenu": item.subMenu} for item in self._items]

    __call__ = getItems


class WebFilterAdmin(CMFSite, MainConfigFile):
    """webfilter admin object"""

    addPermission = "Add WebFilterAdmin Object"
    ourServices = ("squid", "dansguardian", "postgresql", systemService)
    meta_type = "WebFilterAdmin"
    configDir = "/etc/dansguardian"
    security = ClassSecurityInfo()
    skin = "Base"
    skinBasePath = "custom, proxyadmin_base, proxyadmin_static, " \
                   "proxyadmin_users, proxyadmin_%s, proxyadmin_db, " \
                   "zpt_topic, zpt_content, zpt_generic, zpt_control, Images"
    
    def __init__(self, id, title="", userFolder=True):
        CMFSite.__init__(self, id, title)

        
        MainConfigFile.__init__(self, id, title,
                                os.path.join(self.configDir, "dansguardian.conf"),
                                os.path.join(self.configDir, "dansguardianf1.conf"))
        
    security.declarePrivate("setupWebFilterAdmin")
    def setupWebFilterAdmin(self):
        """webfilteradmin setup"""
        
        self.manage_addProduct['CMFCalendar'].manage_addTool('CMF Calendar Tool',
                                                             None)
        skinsTool = getToolByName(self, 'portal_skins')
        addDirectoryViews(skinsTool, "skins", globals())
        # FIXME: why psycopg? :-)
        skinsTool.addSkinSelection(self.skin, self.skinBasePath % 'psycopg',
                                   make_default=1)
        self.setupCurrentSkin()

        self.manage_permission(ManageWebFilter, ['Member', 'Manager'], 0)
        self.manage_permission(SetOwnPassword, ['Member', 'Manager'],   1)
        self.manage_permission(AddPortalMember, ['Manager'], 0)
        
        membershipTool = getToolByName(self, 'portal_membership')
        membershipTool.setMemberareaCreationFlag()

        #actionsTool = getToolByName(self, 'portal_actions')

        #section = None
        #for action in webFilterAdminActions:
        #    actionsTool.addAction( XXX )

    _admin_actions = Menu(
        ('Ammistrazione', 'admin'),
        ('Eventi', 'events'),
        ('Siti permessi', 'exceptionsitelist/',
         ('URL permessi', 'exceptionurllist/'),
         ('IP permessi', 'exceptioniplist/'),
         ),
        ('Siti bloccati', 'bannedsitelist/',
         ('Estensioni bloccate', 'bannedextensionlist/'),
         ('Tipi MIME bloccati', 'banned<mimetypelist/'),
         ('URL bloccati', 'bannedurllist/'),
#         ('Regexp URL bloccati', 'bannedregexpurllist/'),
         ('IP bloccati', 'bannediplist/')
         ),
        ('Statistiche server', 'statistics/server',
         ('Settimanali', 'statistics/server/week'),
         ('Mensili', 'statistics/server/month'),
         ('Annuali', 'statistics/server/year')
         ),
        ('Statistiche client', 'statistics/client',
         ('Settimanali', 'statistics/client/week'),
         ('Mensili', 'statistics/client/month'),
         ('Annuali', 'statistics/client/year')
         ),
        ('Cerca nei log', 'search'),
        )

    _anonymous_actions = Menu(
        ('Login', 'login_form'),
        )

    security.declarePublic('actions')
    def getMenuActions(self):
        """default admin actions"""

        base = self.absolute_url()
        if not self.isAuthenticated():
            return self._anonymous_actions.getItems(base)
        return self._admin_actions.getItems(base)

    security.declarePublic('isAuthenticated')
    def isAuthenticated(self):
        """isAuthenticated"""

        return not getSecurityManager().getUser().getId() is None

    security.declareProtected(ManageWebFilter, "supportedConnectionTypes")
    def supportedConnectionTypes(self):
        """supported sql connection types"""

        return dict(PostgreSQL=("ZPsycopgDA", "manage_addZPsycopgConnection"),
                    SQLite=("ZSQLiteDA", "manage_addZSQLiteConnection"),
                    MySQL=("ZMySQLDA", "manage_addZMySQLConnection"))

    security.declareProtected(ManageWebFilter, "getConnectionType")
    def getConnectionType(self):
        """get db connection"""

        try:
            return self.dbConnection.title
        except AttributeError:
            return ''

    security.declareProtected(ManageProperties, "editConfig")
    def editConfig(self, dbType=None, dbParameters=''):
        """edit config"""

        # - salva tutti i parametri nell'istanza: hostname/ip, tipoconnessione,
        #   parametriconnessione.
        # - genera /etc/dansguardian/template.html dal file XXX,
        #   con i parametri YYY.
        # - rigenera il dbConnection in base ai parametri ZZZ.

        #self.hostname = hostname
        if dbType:
            try:
                self.manage_delObjects(ids='dbConnection')
            except BadRequest:
                pass
            parms = self.supportedConnectionTypes()[dbType]
            creator = getattr(self.manage_addProduct[parms[0]],
                              parms[1])
            creator('dbConnection', dbType, dbParameters)
        else:
            dbType = self.getConnectionType()

        skinsTool = self.portal_skins
        skinsTool.manage_skinLayers(chosen=(self.skin, ),
                                    del_skin=True)
        skinsTool.addSkinSelection(self.skin,
                                   self.skinBasePath % dbType.lower(),
                                   make_default=1)
        return self.config(message="Configurazione salvata (%r)"  % (dbType, ))
    
    security.declareProtected(ManageProperties, "editStaticConfig")
    def editStaticConfig(self, serverHostname):
        """edit static config"""

        self.serverHostname = serverHostname

    security.declareProtected(View, "index_html")
    def index_html(self, REQUEST=None):
        """index_html"""

        REQUEST.RESPONSE.redirect(self.getMenuActions()[0]['url'])
        return ''

    # utils...

    security.declarePublic('formatDate')
    def formatDate(self, date):
        """format date"""

        if isinstance(date, basestring):
            date = DateTime(date)
        return date.strftime("%d/%m/%Y %H:%M")
    
    security.declarePublic('formatDateShort')
    def formatDateShort(self, date=None):
        """format date short"""

        if isinstance(date, int):
            date = DateTime() + date
        elif date is None:
            date = DateTime()
        return date.strftime("%d/%m/%Y")

    security.declareProtected(ManageWebFilter, "serviceControl")
    def serviceControl(self, start=None, stop=None, reload=None, REQUEST=None):
        """service control"""

        if start in self.ourServices:
            action = "start"
            service = start
        elif stop in self.ourServices:
            action = "stop"
            service = stop
        elif reload in self.ourServices:
            action = "reload"
            service = reload
        else:
            action = None
        
        if action:
            if Service(service).serviceControl(action):
                msg = "%s del servizio %s eseguito con successo." % \
                      (action, service)
            else:
                msg = "Errore durante la fase di %s del servizio %s." % \
                      (action, service)
        else:
            msg = ""
        if not REQUEST is None:
            return self.admin(message=msg)
    
    security.declareProtected(View, "listServices")
    def listServices(self):
        """list proxy processes

        return a dictionary proxy-process: boolean(running/non-running)
        """

        services = []
        for name in self.ourServices:
            service = Service(name)
            services.append({"name": service.getName(),
                             "running": service.running()})
        return services

    security.declarePrivate("dansguardianReload")
    def dansguardianReload(self):
        """reload dansguardian"""
        
        return Service("dansguardian").serviceControl("reload")
    
    security.declareProtected(ManageWebFilter, "processEvents")
    def processEvents(self, ids=(), addToBanned=None, addToException=None,
                      ignore=None, REQUEST=None):
        """process ids and action for events"""
        
        if not ids or not isinstance(ids, (tuple, list)):
            return self.events(message="Selezionare almeno un elemento della lista.")

        if addToException:
            error = self.exceptionsitelist.addNotifiedList(ids)
            if error:
                return self.events(message=error)
        elif addToBanned:
            error = self.bannedsitelist.addNotifiedList(ids)
            if error:
                return self.events(message=error)
        elif ignore:
            pass
        else:
            return self.events(message="Nessuna azione eseguita.")
        self.eraserequestedUrls(requestid=", ".join(ids))
        return self.events(message="Operazione completata.")
    
    security.declareProtected(View, "requestAddToException")
    def requestAddToException(self, req_url, req_client, req_message,
                              req_reasonlogged,
                              REQUEST=None):
        """requestAddToException"""

        self.addToRequest(url=req_url,
                          clientName=getfqdn(req_client),
                          message=req_message,
                          detail=req_reasonlogged)
        return self.notifyOk()


InitializeClass(WebFilterAdmin)


class WebFilterAdminGenerator(PortalGenerator):
    """
    webfilteradmin generator
    """

    klass = WebFilterAdmin

InitializeClass(WebFilterAdminGenerator)

def manage_addWebFilterAdmin(self, id, title="WebFilterAdmin", description="",
                             create_userfolder=1,
                             email_from_address="postmaster@localhost",
                             email_from_name="WebFilter Administrator",
                             validate_email=0, default_charset='',
                             RESPONSE=None):
    """ Adds a webfilteradmin(portal) instance.
    """
    gen = WebFilterAdminGenerator()
    id = id.strip()
    p = gen.create(self, id, create_userfolder)
    gen.setupDefaultProperties(p, title, description,
                               email_from_address, email_from_name,
                               validate_email, default_charset)
    p.setupWebFilterAdmin()
    if RESPONSE is not None:
        RESPONSE.redirect(p.absolute_url() + "/finish_portal_construction")

manage_addWebFilterAdminForm = PageTemplateFile( \
    "www/addWebFilterAdminForm", globals())

