#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# parseguardian is dansguardian log parser to sql/db.
# Copyright (C) 2005 Marco Andreini
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
parseguardian: dansguardian log parser to sql.

"""

import sys
import os
import logging
import ZConfig

from copy import deepcopy
from time import strptime
from pwd import getpwnam

from sqllog import dbLogConnect

__version__ = 0
__revision__ = 9
__author__ = "Marco Andreini"
__email__ = "m.andreini(at)tiscali.it"

configFileName = "parseguardian.conf"
logging.basicConfig()
log = logging.getLogger("parseguardian")


class LogParser:
    """logParser(filename)

    Note: '-' is stdin
    """

    __denied = "*DENIED* "
    __deniedLen = len(__denied)
    
    def __init__(self, filename):
        if filename == "-":
            self.file = sys.stdin
        else:
            if not os.path.exists(filename):
                # WARNING: others check on log-file? is regular-file?
                log.critical("log-file %r not found (fifo).", filename)
            self.file = open(filename, "r")
    
    def parsing(self, insertFunction, separator, messages):
        """parse log file by line"""
        
        while 1:
            line = self.file.readline()
            if not line:
                # eof
                break
            # remove end-of-line
            if line.endswith("\n"):
                line = line[:-1]
            if not line:
                # ignore empty lines
                continue
            
            # parse the line
            try:
                date, dash, source, url, message, method, size = \
                      line.split(separator, 7)
            except ValueError:
                log.error("parsing error on log (line %r)" % line)
                continue
            
            message = message.strip()
            if message.startswith(self.__denied):
                reason = messages.getCode(message[self.__deniedLen:])
            else:
                # FIXME: treat *EXCEPT*
                reason = 0
            
            # WARNING: content of message is important? ... discarded.
            date = strptime(date, "%Y.%m.%d %H:%M:%S")

            if dash != "-":
                log.error("missing \"-\" on log (line %r)" % line)
            size = int(size)
            
            insertFunction(date, source, url, method, size, reason)


class Messages:
    """dansguardian-messages(filename)

    Fast lookup for message strings.
    """
    
    def __init__(self, filename):
        messages = []
        lmin = 0
        for line in open(filename, "r").xreadlines():
            line = line.strip()
            if not line or line.startswith("#"):
                # ignore comments and empty lines
                continue

            # FIXME: too pythonish...
            code, message = eval(line, {})
            if lmin == 0 or len(message) < lmin:
                # minimum message length
                lmin = len(message)
            messages.append((message, int(code)))
        
        # process messages & codes
        msgs = {}
        for message, code in messages:
            starts = message[:lmin]
            if not msgs.has_key(starts):
                msgs[starts] = []
            msgs[starts].append((message[lmin:], code))
        # reverse-sort all messages
        for value in msgs.values():
            value.sort()
            value.reverse()
        
        # deepcopy for fast lookup
        self._messages = deepcopy(msgs)
        self._minLength = lmin

        # DEBUG:
        #from pprint import pformat
        #print pformat(self._messages)
        
        log.debug("Parsed and hashed %d messages from %r." % \
                  (len(messages), filename))
    
    def getCode(self, message):
        """get code from message"""
        
        if len(message) < self._minLength:
            log.error("invalid message length: %r" % (message, ))
        
        startswith = message[:self._minLength]
        try:
            for rest, code in self._messages[startswith]:
                if message.startswith(rest, self._minLength):
                    # Note: get first result.
                    return code
        except KeyError:
            log.warn("unknown message: %r" % (message, ))
        return 0


def main():
    """main function"""

    progdir = os.path.dirname(sys.argv[0])
    schema = ZConfig.loadSchema(os.path.join(progdir, "schema.xml"))
    config, handler = ZConfig.loadConfig(schema,
                                         os.path.join(progdir, configFileName))
    
    level = config.verbose
    if level not in ("debug", "info", "warning", "error"):
        level = "info"
    log.setLevel(getattr(logging, level.upper()))
    
    log.info("PID=%d", os.getpid())
    
    if config.effectiveuser:
        uid, gid = getpwnam(config.effectiveuser)[2:4]
        try:
            os.setgid(gid)
            os.setuid(uid)
        except OSError:
            log.error("error setuid to %r, ignored." % (config.effectiveuser, ))
    
    logParser = LogParser(config.logfilename)
    
    if config.usepsyco:
        try:
            import psyco
            psyco.profile()
        except ImportError:
            log.error("error importing psyco.")
    
    messages = Messages(config.messagesfile)
    sqlConnectionString = config.sqlconnection
    sqlType = config.sqlserver.lower()
    dbLog = dbLogConnect(sqlType, sqlConnectionString)

    log.info("start processing log.")
    logParser.parsing(dbLog.insert, eval(config.logseparator, {}), messages)

    dbLog.close()
    log.info("end processing log.")


if __name__ == "__main__":
    main()

