# -*- coding: utf-8 -*-
#
# parseguardian is dansguardian log parser to sql/db.
# Copyright (C) 2005 Marco Andreini
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
sql interface for parseguardian.

"""

__version__ = 0
__revision__ = 9
__author__ = "Marco Andreini"
__email__ = "m.andreini(at)tiscali.it"


import logging
from socket import getfqdn
from time import strftime, mktime
from urlparse import urlparse
from random import randrange


log = logging.getLogger("parseguardian")


class SQLUniqueValues:
    """
    sql table with unique values (column)

    """

    queries = dict(
        select="SELECT %(table)sId FROM %(table)s" \
        " WHERE %(column)s = '%%s'",
        # fail on empty table
        next="SELECT MAX(%(table)sId) + 1 FROM %(table)s",
        insert="INSERT INTO %(table)s (%(table)sId, %(column)s)" \
        " VALUES (%%d, '%%s')"
        )

    def __init__(self, cursor, table, column):
        queries = self.queries
        self.selectQuery = queries["select"] % locals()
        self.nextQuery = queries["next"] % locals()
        self.insertQuery = queries["insert"] % locals()
        self.cursor = cursor
    
    def insert(self, value):
        """insert key value if needed and get value"""

        cursor = self.cursor
        cursor.execute(self.selectQuery % (value, ))
        result = cursor.fetchone()
        if result:
            return result[0]
        # get next id & insert value
        
        cursor.execute(self.nextQuery)
        result = cursor.fetchone()
        rid = result[0]
        result = cursor.execute(self.insertQuery % (rid, value))
        return rid


class PgsqlUniqueValues(SQLUniqueValues):
    """pgsql table with unique values (column)"""

    # table_tableid_seq convention
    queries = SQLUniqueValues.queries.copy()
    queries.update(dict(
        next="SELECT NEXTVAL('%(table)s_%(table)sid_seq')"
        ))


class MysqlUniqueValues(SQLUniqueValues):
    """mysql table with unique values (column)"""
    
    queries = SQLUniqueValues.queries.copy()
    queries.update(dict(
        next="SELECT LAST_INSERT_ID()",
        insert="INSERT INTO %(table)s (%(table)sId, %(column)s)" \
        " VALUES (NULL, '%%s')"
        ))
    
    def insert(self, value):
        """insert key value if needed and get value"""

        cursor = self.cursor
        cursor.execute(self.selectQuery % (value, ))
        result = cursor.fetchone()
        if result:
            return result[0]
        # insert value & get inserted id
        
        cursor.execute(self.insertQuery % (value, ))
        result = cursor.execute(self.nextQuery)
        result = cursor.fetchone()
        return result[0]


class SQLiteUniqueValues(MysqlUniqueValues):
    """SQLite table with unique values (column)"""

    queries = MysqlUniqueValues.queries.copy()
    queries.update(dict(
        next="SELECT last_insert_rowid()"
        ))


class ResolvedValues:
    """fqdn resolved value (~address)"""
    
    def __init__(self, parent):
        self.parent = parent
    
    def insert(self, value):
        """insert"""
        
        return self.parent.insert(getfqdn(value))


class UniqueValuesCached:
    """
    unique-values-cached (parent, maxsize)

    """
    
    def __init__(self, parent, maxsize):
        self.maxsize = maxsize
        # circular buffer
        self.circular = []
        self.data = {}
        self.parent = parent

        #self.st_hit = 0
        #self.st_count = 0
    
    def insert(self, value):
        """cached insert

        if size == maxsize a randomly choosed item is removed.
        
        """
        
        circular = self.circular
        data = self.data

        if not data.has_key(value):
            if len(circular) >= self.maxsize:
                position = randrange(self.maxsize)
                del data[circular[position]]
                circular[position] = value
            else:
                circular.append(value)
            
            res = self.parent.insert(value)
            data[value] = res
            return res

        return data[value]


class SQLLog:
    """
    generic sql queries to dblog
    
    """
    
    queries = dict(
        selectLastDate="SELECT MAX(date) FROM log",
        # WARNING: auto increment is not generic...
        insertIntoLog="INSERT INTO log VALUES (NULL," \
        " %s, %d, %d, %d, %s, %d, %d, %d, %d)"
        )

    clientCacheSize = 128
    serverCacheSize = 512
    protocolCacheSize = 4
    methodCacheSize = 4
    insertCacheSize = 64

    def __init__(self, connection, SQLUniqueValues):
        self.connection = connection
        cursor = connection.cursor()
        self.cursor = cursor

        self.selectLastDate = self.queries["selectLastDate"]
        self.insertIntoLog = self.queries["insertIntoLog"]
        
        self.clients = UniqueValuesCached(
            ResolvedValues(SQLUniqueValues(cursor,
                                           table="client",
                                           column="hostname")),
            self.clientCacheSize)
        self.servers = UniqueValuesCached(
            SQLUniqueValues(cursor,
                            table="server",
                            column="hostname"),
            self.serverCacheSize)
        self.protocols = UniqueValuesCached(
            SQLUniqueValues(cursor,
                            table="protocol",
                            column="name"),
            self.protocolCacheSize)
        self.methods = UniqueValuesCached(
            SQLUniqueValues(cursor,
                            table="method",
                            column="name"),
            self.methodCacheSize)
        
        cursor.execute(self.selectLastDate)
        result = cursor.fetchone()
        if result and result[0]:
            self.date = result[0].strftime("%Y-%m-%d %H:%M:01")
            log.info("Starting from log line after %r." % (self.date, ))
        else:
            self.date = 0
        self.counter = 0

    def getIds(self, source, destination, protocol, method):
        """get ids required for next insert"""

        return (self.clients.insert(source), self.servers.insert(destination),
                self.protocols.insert(protocol), self.methods.insert(method))
    
    def nextInsert(self, date, clientId, serverId, protocolId, path, methodId,
                   size, reason, count):
        """nextInsert"""
        
        self.cursor.execute(self.insertIntoLog,
                            (date, clientId, serverId, protocolId, path,
                             methodId, size, reason, count))
        self.counter += 1
        if self.counter >= self.insertCacheSize:
            self.counter = 0
            self.connection.commit()
    
    def firstInsert(self, date, *args, **kw):
        """firstInsert (skip old entries)"""
        
        if date >= self.date:
            if self.counter:
                log.info("Skip %d old entries." %  (self.counter + 1, ))
            self.insert = self.nextInsert
            self.insert(date, *args, **kw)
        self.counter += 1
    
    insert = firstInsert
    
    def close(self):
        self.connection.commit()
        self.connection.close()


class PgsqlDBLog(SQLLog):
    """postgresql dblog"""

    queries = SQLLog.queries.copy()
    queries.update(dict(
        insertIntoLog="INSERT INTO log VALUES (NEXTVAL('log_logid_seq')," \
        " %s, %d, %d, %d, %s, %d, %d, %d, %d)"
        ))


class MysqlDBLog(SQLLog):
    """mysql dblog"""

    # NULL for newer insert (now is in parent class).
    queries = SQLLog.queries.copy()


class SQLiteDBLog(SQLLog):
    """sqlite dblog"""

    # SQLite MAX(date) return float.
    queries = SQLLog.queries.copy()
    queries.update(dict(
        selectLastDate="SELECT date FROM log ORDER BY date DESC LIMIT 1"
        ))


class DBLogAntiFlood:
    """
    dblog with anti-flood
    """
    
    # interval, in seconds: the time granularity.
    interval = 60.0
    
    def __init__(self, dblog):
        self.dblog = dblog
        self.data = {}
        self.tnext = None
    
    def insert(self, date, source, url, method, size, reason):
        """insert data into log (or skip to next)"""
        
        tdate = mktime(date)
        sdate = strftime("%Y-%m-%d %H:%M:00", date)
        
        protocol, destination, path, dummy1, parameters, dummy2 = urlparse(url)
        clientId, serverId, protocolId, methodId = \
                  self.dblog.getIds(source, destination, protocol, method)
        
        args = (sdate, clientId, serverId, protocolId, path, methodId,
                size, reason)
        try:
            self.data[args] += 1
        except KeyError:
            # WARNING: there is no max-size for self.data, frequently flushed.
            self.data[args] = 1
        
        if self.tnext is None:
            self.tnext = tdate + self.interval
        elif tdate >= self.tnext:
            self.tnext = tdate + self.interval
            self.flush()
    
    def flush(self):
        """flush cache"""
        
        for args, count in self.data.items():
            self.dblog.insert(*(args + (count, )))
        self.data = {}
    
    def close(self):
        """close connection"""
        
        self.flush()
        self.dblog.close()


def dbLogConnect(sqlType, sqlConnectionString):
    """dbLogConnect"""

    if sqlType == "postgresql":
        import psycopg
        
        dblog = PgsqlDBLog(psycopg.connect(sqlConnectionString),
                           PgsqlUniqueValues)
    elif sqlType == "mysql":
        import MySQLdb
        
        # (simple) decoding of mysql-connection-string
        kwargs = {}
        for item in sqlConnectionString.split(" "):
            key, value = item.strip().split("=", 1)
            kwargs[key] = value
        dblog = MysqlDBLog(MySQLdb.connect(**kwargs),
                           MysqlUniqueValues)
    elif sqlType == "sqlite":
        import sqlite
        
        dblog = SQLiteDBLog(sqlite.connect(sqlConnectionString),
                            SQLiteUniqueValues)
    else:
        log.critical("unknown sqlType %r" % (sqlType, ))
        raise ValueError, "sqlType error: %r" % (sqlType, )
    
    return DBLogAntiFlood(dblog)


